import asyncio
import codecs
import functools
import time

from .. import fhem, generic, utils
from ..core import ble

DEFAULT_TIMEOUT = 1


class blue_connect(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self._ble_lock = asyncio.Lock()
        self._conn = None
        self.water_temp = "0"
        self.water_orp = "0"
        self.water_ph = "0"
        self.other_18 = ""
        set_conf = {
            "measure": {"help": "Send signal to start measuring"},
        }
        self.set_set_config(set_conf)
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) != 4:
            return "Usage: define my_blueconnect fhempy blue_connect MAC"
        self._mac = args[3]
        self.hash["MAC"] = self._mac
        self._conn = ble.BTLEConnection(
            self._mac,
            keep_connected=True,
        )
        self._conn.set_callback("all", self.received_notification)
        self.create_async_task(self.update_loop())
        self.create_async_task(self.keep_connected())

    async def Undefine(self, hash):
        if self._conn:
            self._conn.set_keep_connected(False)
        return await super().Undefine(self.hash)

    async def set_measure(self, hash, params):
        self.create_async_task(self.measure_once())

    def received_notification(self, data):
        self.raw_measurement = codecs.encode(data, "hex")
        raw_temp = int(self.raw_measurement[4:6] + self.raw_measurement[2:4], 16)
        self.water_temp = float(raw_temp) / 100

        raw_ph = int(self.raw_measurement[8:10] + self.raw_measurement[6:8], 16)
        self.water_ph = round((float(0x0800) - float(raw_ph)) / 232 + 7, 2)

        raw_orp = int(self.raw_measurement[12:14] + self.raw_measurement[10:12], 16)
        self.water_orp = round(float(raw_orp) / 4)

        raw_battery = int(self.raw_measurement[20:22] + self.raw_measurement[18:20], 16)
        self.battery = float(raw_battery) / 1000

    def blocking_measure(self):
        for cnt in range(0, 10):
            try:
                # enable notifications
                self._conn.write_characteristic(0x0014, b"\x01\x00")
                # start measuring
                self._conn.write_characteristic(0x0012, b"\x01", 60)
                break
            except BrokenPipeError:
                self.logger.error("BrokenPipeError, reconnect BLE")
                self._conn.stop_helper()
            except Exception:
                self.logger.exception("Failed to write characteristics")
                time.sleep(10)

    def blocking_read_others(self):
        for cnt in range(0, 10):
            try:
                data = self._conn.read_characteristic(0x18)
                self.other_18 = codecs.encode(data, "hex")
            except BrokenPipeError:
                self.logger.error("BrokenPipeError, reconnect BLE")
                self._conn.stop_helper()
            except Exception:
                self.logger.exception("Failed to read characteristics")
                time.sleep(10)

    async def update_loop(self):
        while True:
            try:
                await self.measure_once()
            except asyncio.CancelledError:
                break
            except Exception:
                self.logger.exception("Failed to update readings")
            await asyncio.sleep(7200)

    async def keep_connected(self):
        while True:
            try:
                async with self._ble_lock:
                    await utils.run_blocking(
                        functools.partial(self.blocking_read_others)
                    )
            except asyncio.CancelledError:
                break
            except Exception:
                self.logger.exception("Failed to update readings")
            await asyncio.sleep(60)

    async def measure_once(self):
        self.water_temp = 0
        self.water_orp = 0
        self.water_ph = 0
        self.battery = 0

        async with self._ble_lock:
            await utils.run_blocking(functools.partial(self.blocking_measure))
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdate(self.hash, "unknown_handle_18", self.other_18)
        await fhem.readingsBulkUpdate(self.hash, "temperature", self.water_temp)
        await fhem.readingsBulkUpdate(self.hash, "ph", self.water_ph)
        await fhem.readingsBulkUpdate(self.hash, "orp", self.water_orp)
        await fhem.readingsBulkUpdate(self.hash, "raw", self.raw_measurement)
        await fhem.readingsBulkUpdate(self.hash, "battery", self.battery)

        state = []
        if self.water_ph < 7.2:
            state.append(
                f"Ph: {self.water_ph} too low, diff {round(7.2-self.water_ph, 2)}"
            )
            await fhem.readingsBulkUpdate(self.hash, "ph_state", "low")
        elif self.water_ph > 7.7:
            state.append(
                f"Ph: {self.water_ph} too high, diff {round(7.7-self.water_ph, 2)}"
            )
            await fhem.readingsBulkUpdate(self.hash, "ph_state", "high")
        else:
            await fhem.readingsBulkUpdate(self.hash, "ph_state", "ok")

        if self.water_orp < 550:
            state.append(
                f"ORP: {self.water_orp} too low, diff {round(550-self.water_orp)}"
            )
            await fhem.readingsBulkUpdate(self.hash, "orp_state", "low")
        elif self.water_orp > 650:
            state.append(
                f"ORP: {self.water_orp} too high, diff {round(650-self.water_orp)}"
            )
            await fhem.readingsBulkUpdate(self.hash, "orp_state", "high")
        else:
            await fhem.readingsBulkUpdate(self.hash, "orp_state", "ok")

        if len(state) == 0:
            state = ["ok"]

        await fhem.readingsBulkUpdate(self.hash, "state", ", ".join(state))

        await fhem.readingsEndUpdate(self.hash, 1)



import asyncio
import janus
from fhempy.lib.ble_monitor.bt_helpers import (
    BT_INTERFACES,
    DEFAULT_BT_INTERFACE,
    DEFAULT_HCI_INTERFACE,
)
from fhempy.lib.ble_monitor.hcidump import HCIdump


from .const import (
    CONF_DEVICE_ENCRYPTION_KEY,
    CONF_DISCOVERY,
    CONF_MAC,
    CONF_DEVICES,
    CONF_ACTIVE_SCAN,
    CONF_BATT_ENTITIES,
    CONF_BT_AUTO_RESTART,
    CONF_BT_INTERFACE,
    CONF_DECIMALS,
    CONF_HCI_INTERFACE,
    CONF_GATEWAY_ID,
    CONF_PERIOD,
    CONF_LOG_SPIKES,
    CONF_REPORT_UNKNOWN,
    CONF_RESTORE_STATE,
    CONF_USE_MEDIAN,
    CONF_UUID,
    DEFAULT_ACTIVE_SCAN,
    DEFAULT_BATT_ENTITIES,
    DEFAULT_BT_AUTO_RESTART,
    DEFAULT_DECIMALS,
    DEFAULT_DISCOVERY,
    DEFAULT_LOG_SPIKES,
    DEFAULT_PERIOD,
    DEFAULT_REPORT_UNKNOWN,
    DEFAULT_RESTORE_STATE,
    DEFAULT_USE_MEDIAN,
)


class BLEmonitor:
    """BLE scanner."""

    __instance = None

    @staticmethod
    def getInstance(logger):
        """Static access method."""
        if BLEmonitor.__instance is None:
            BLEmonitor.__instance = BLEmonitor(logger)
        return BLEmonitor.__instance

    def __init__(self, logger):
        """Init."""
        self.logger = logger
        self.dataqueue = {
            "binary": janus.Queue(),
            "measuring": janus.Queue(),
            "tracker": janus.Queue(),
        }
        self.config = {
            CONF_DEVICES: [],
            CONF_ACTIVE_SCAN: DEFAULT_ACTIVE_SCAN,
            CONF_BATT_ENTITIES: DEFAULT_BATT_ENTITIES,
            CONF_BT_AUTO_RESTART: DEFAULT_BT_AUTO_RESTART,
            CONF_BT_INTERFACE: [],
            CONF_DECIMALS: DEFAULT_DECIMALS,
            CONF_HCI_INTERFACE: [],
            CONF_GATEWAY_ID: "",
            CONF_PERIOD: DEFAULT_PERIOD,
            CONF_LOG_SPIKES: DEFAULT_LOG_SPIKES,
            CONF_REPORT_UNKNOWN: DEFAULT_REPORT_UNKNOWN,
            CONF_RESTORE_STATE: DEFAULT_RESTORE_STATE,
            CONF_USE_MEDIAN: DEFAULT_USE_MEDIAN,
            CONF_UUID: "",
            CONF_DISCOVERY: DEFAULT_DISCOVERY,
        }
        self.dumpthread = None
        self.fhem_devices = {}
        self.receive_from_queues()

    def receive_from_queues(self):
        self.task_measuring = asyncio.create_task(self.receive_from_measuring())
        self.task_tracker = asyncio.create_task(self.receive_from_tracker())

    async def receive_from_measuring(self):
        while True:
            try:
                measuring = await self.dataqueue["measuring"].async_q.get()
                if "mac" in measuring and measuring["mac"].lower() in self.fhem_devices:
                    for fhem_dev in self.fhem_devices[measuring["mac"].lower()]:
                        await fhem_dev.received_data(measuring)
            except Exception:
                self.logger.exception("Failed to receive_from_measuring")
                asyncio.sleep(10)

    async def receive_from_tracker(self):
        while True:
            try:
                tracker = await self.dataqueue["tracker"].async_q.get()
                if "mac" in tracker and tracker["mac"].lower() in self.fhem_devices:
                    for fhem_dev in self.fhem_devices[tracker["mac"].lower()]:
                        await fhem_dev.received_data(tracker)
            except Exception:
                self.logger.exception("Failed to receive_from_tracker")
                asyncio.sleep(10)

    def register_device(self, fhemdevice):
        simple_mac = fhemdevice.mac().replace(":", "").lower()
        if simple_mac not in self.fhem_devices:
            self.fhem_devices[simple_mac] = []
        self.fhem_devices[simple_mac].append(fhemdevice)
        self.config[CONF_DEVICES].append(
            {
                CONF_MAC: fhemdevice.mac(),
                CONF_DEVICE_ENCRYPTION_KEY: fhemdevice.encryption_key(),
            }
        )

        self.update_hci_interface(fhemdevice.hci())

        self.restart()

    def update_hci_interface(self, intf):
        self.config[CONF_HCI_INTERFACE] = []
        self.config[CONF_BT_INTERFACE] = []
        for fhem_mac in self.fhem_devices:
            for fhem_dev in self.fhem_devices[fhem_mac]:
                self.config[CONF_HCI_INTERFACE].append(fhem_dev.hci())
                self.config[CONF_BT_INTERFACE].append(BT_INTERFACES[fhem_dev.hci()])

    def unregister_device(self, fhemdevice):
        simple_mac = fhemdevice.mac().replace(":", "").lower()
        try:
            self.fhem_devices[simple_mac].remove(fhemdevice)
        except Exception:
            pass
        try:
            self.config[CONF_DEVICES].remove(
                {
                    CONF_MAC: fhemdevice.mac(),
                    CONF_DEVICE_ENCRYPTION_KEY: fhemdevice.encryption_key(),
                }
            )
        except Exception:
            pass

        self.update_hci_interface(fhemdevice.hci())

        self.restart()

    def shutdown_handler(self, event):
        """Run homeassistant_stop event handler."""
        self.logger.debug("Shutdown event fired: %s", event)
        self.stop()

    def start(self):
        """Start receiving broadcasts."""
        self.logger.debug("Spawning HCIdump thread")
        self.dumpthread = HCIdump(
            self.logger,
            config=self.config,
            dataqueue=self.dataqueue,
        )
        self.dumpthread.start()

    def stop(self):
        """Stop HCIdump thread(s)."""
        result = True
        if self.dumpthread is None:
            self.logger.debug("BLE monitor stopped")
            return True
        if self.dumpthread.is_alive():
            self.dumpthread.join()
            if self.dumpthread.is_alive():
                result = False
                self.logger.error(
                    "Waiting for the HCIdump thread to finish took too long! (>10s)"
                )
        self.logger.debug("BLE monitor stopped")
        return result

    def restart(self):
        """Restart scanning."""
        # wait 3s for other restarts and restart only once
        self.stop()
        self.start()



import functools
import logging
import re
import traceback
from datetime import datetime

from ..utils import run_blocking
from .const import (
    COUNT_FORCE_UPDATE_AFTER_COMMAND,
    DATE_FORMAT,
    INITIAL_STATUS_DELAY_AFTER_COMMAND,
    INTERVAL_FORCE_UPDATE_AFTER_COMMAND,
    RECHECK_STATUS_DELAY_AFTER_COMMAND,
    REGION_CANADA,
    REGIONS,
    START_FORCE_UPDATE_AFTER_COMMAND,
    TOPIC_UPDATE,
    VEHICLE_ENGINE_TYPE,
    VEHICLE_LOCK_ACTION,
)
from .KiaUvoApiImpl import KiaUvoApiImpl
from .Token import Token

_LOGGER = logging.getLogger(__name__)


class Vehicle:
    def __init__(
        self,
        token: Token,
        kia_uvo_api: KiaUvoApiImpl,
        unit_of_measurement: str,
        enable_geolocation_entity: bool,
        region: str,
    ):
        self.token: Token = token
        self.kia_uvo_api: KiaUvoApiImpl = kia_uvo_api
        self.unit_of_measurement: str = unit_of_measurement
        self.enable_geolocation_entity: bool = enable_geolocation_entity
        self.region: str = region
        self.name = token.vehicle_name
        self.model = token.vehicle_model
        self.id = token.vehicle_id
        self.registration_date = token.vehicle_registration_date
        self.vehicle_data = {}
        self.engine_type = None
        self.last_updated: datetime = datetime.min
        self.force_update_try_caller = None
        self.topic_update = TOPIC_UPDATE.format(self.id)
        self.current_ev_battery = None
        _LOGGER.debug(f"Received token into Vehicle Object {vars(token)}")

    async def update(self):
        try:
            await self.refresh_token()
            previous_vehicle_status = self.get_child_value("vehicleStatus")
            previous_vehicle_location = self.get_child_value("vehicleLocation")
            self.vehicle_data = await run_blocking(
                functools.partial(
                    self.kia_uvo_api.get_cached_vehicle_status, self.token
                )
            )
            self.set_last_updated()
            self.set_engine_type()
            if self.enable_geolocation_entity:
                await run_blocking(
                    functools.partial(
                        self.set_geocoded_location, previous_vehicle_location
                    )
                )

            if (
                not self.get_child_value("vehicleStatus.engine")
                and previous_vehicle_status is not None
                and not previous_vehicle_status["engine"]
                and self.get_child_value("vehicleStatus.evStatus.batteryStatus") == 0
                and previous_vehicle_status["evStatus"]["batteryStatus"] != 0
            ):
                _LOGGER.debug(
                    "zero battery api error, force_update started to correct data"
                )
                await self.force_update()

        except Exception as ex:
            _LOGGER.error(
                "Exception in update : %s - traceback: %s",
                ex,
                traceback.format_exc(),
            )

    async def force_update(self):
        await run_blocking(
            functools.partial(self.kia_uvo_api.update_vehicle_status, self.token)
        )
        await self.update()

    async def force_update_loop_start(self):
        if self.kia_uvo_api.last_action_tracked:
            # self.force_update_try_caller = async_call_later(
            #    self.hass,
            #    INITIAL_STATUS_DELAY_AFTER_COMMAND,
            #    self.check_action_completed_loop,
            # )
            pass
        else:
            self.force_update_try_count = 0
            # self.force_update_try_caller = async_call_later(
            #    self.hass, START_FORCE_UPDATE_AFTER_COMMAND, self.force_update_loop
            # )

    async def check_action_completed_loop(self, _):
        await run_blocking(
            functools.partial(self.kia_uvo_api.check_last_action_status, self.token)
        )
        if self.kia_uvo_api.last_action_completed:
            self.kia_uvo_api.last_action_xid = None
            await self.update()
        # else:
        #    async_call_later(
        #        self.hass,
        #        RECHECK_STATUS_DELAY_AFTER_COMMAND,
        #        self.check_action_completed_loop,
        #    )

    async def force_update_loop(self, _):
        _LOGGER.debug(
            f"force_update_loop start {self.force_update_try_count} {COUNT_FORCE_UPDATE_AFTER_COMMAND}"
        )
        if self.force_update_try_count == COUNT_FORCE_UPDATE_AFTER_COMMAND:
            self.force_update_try_count = 0
            return

        last_updated: datetime = self.last_updated
        _LOGGER.debug(f"force_update_loop last_updated {last_updated}")

        await self.force_update()
        _LOGGER.debug(
            f"force_update_loop force_update_finished {last_updated} {self.last_updated}"
        )
        if last_updated == self.last_updated:
            self.force_update_try_count = self.force_update_try_count + 1
            # self.force_update_try_caller = async_call_later(
            #    self.hass, INTERVAL_FORCE_UPDATE_AFTER_COMMAND, self.force_update_loop
            # )

    def set_geocoded_location(self, old_vehicle_location):
        old_lat = None
        old_lon = None
        old_geocode = None
        if (
            not old_vehicle_location is None
            and old_vehicle_location.get("coord") is not None
        ):
            old_lat = old_vehicle_location["coord"]["lat"]
            old_lon = old_vehicle_location["coord"]["lon"]
            old_geocode = old_vehicle_location.get("geocodedLocation", None)

        new_lat = self.get_child_value("vehicleLocation.coord.lat")
        new_lon = self.get_child_value("vehicleLocation.coord.lon")

        if self.vehicle_data.get("vehicleLocation") is None:
            self.vehicle_data["vehicleLocation"] = {}

        if (old_lat != new_lat or old_lon != new_lon) or old_geocode is None:
            self.vehicle_data["vehicleLocation"][
                "geocodedLocation"
            ] = self.kia_uvo_api.get_geocoded_location(new_lat, new_lon)
        else:
            self.vehicle_data["vehicleLocation"]["geocodedLocation"] = old_geocode

    async def lock_action(self, action: VEHICLE_LOCK_ACTION):
        await run_blocking(
            functools.partial(self.kia_uvo_api.lock_action, self.token, action.value)
        )
        await self.force_update_loop_start()

    async def refresh_token(self):
        _LOGGER.debug(
            f"Refresh token started {self.token.valid_until} {datetime.now()} {self.token.valid_until <= datetime.now().strftime(DATE_FORMAT)}"
        )
        if self.token.valid_until <= datetime.now().strftime(DATE_FORMAT):
            _LOGGER.debug("Refresh token expired")
            await run_blocking(functools.partial(self.login))
            return True
        return False

    async def start_climate(self, set_temp, duration, defrost, climate, heating):
        if set_temp is None:
            set_temp = 21
        if duration is None:
            duration = 5
        if defrost is None:
            defrost = False
        if climate is None:
            climate = True
        if heating is None:
            heating = False
        if (
            self.engine_type == VEHICLE_ENGINE_TYPE.EV
            and REGIONS[self.region] == REGION_CANADA
        ):
            await run_blocking(
                functools.partial(
                    self.kia_uvo_api.start_climate_ev,
                    self.token,
                    set_temp,
                    duration,
                    defrost,
                    climate,
                    heating,
                )
            )
        else:
            await run_blocking(
                functools.partial(
                    self.kia_uvo_api.start_climate,
                    self.token,
                    set_temp,
                    duration,
                    defrost,
                    climate,
                    heating,
                )
            )
        await self.force_update_loop_start()

    async def stop_climate(self):
        if (
            self.engine_type == VEHICLE_ENGINE_TYPE.EV
            and REGIONS[self.region] == REGION_CANADA
        ):
            await run_blocking(
                functools.partial(self.kia_uvo_api.stop_climate_ev, self.token)
            )
        else:
            await run_blocking(
                functools.partial(self.kia_uvo_api.stop_climate, self.token)
            )
        await self.force_update_loop_start()

    async def start_charge(self):
        await run_blocking(functools.partial(self.kia_uvo_api.start_charge, self.token))
        await self.force_update_loop_start()

    async def stop_charge(self):
        await run_blocking(functools.partial(self.kia_uvo_api.stop_charge, self.token))
        await self.force_update_loop_start()

    async def set_charge_limits(self, ac_limit: int, dc_limit: int):
        if ac_limit is None:
            ac_limit = 90
        if dc_limit is None:
            dc_limit = 90
        await run_blocking(
            functools.partial(
                self.kia_uvo_api.set_charge_limits, self.token, ac_limit, dc_limit
            )
        )
        await self.force_update_loop_start()

    def login(self):
        self.token = self.kia_uvo_api.login()

    def set_last_updated(self):
        m = re.match(
            r"(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})",
            self.vehicle_data["vehicleStatus"]["time"],
        )
        local_timezone = self.kia_uvo_api.get_timezone_by_region()
        last_updated = datetime(
            year=int(m.group(1)),
            month=int(m.group(2)),
            day=int(m.group(3)),
            hour=int(m.group(4)),
            minute=int(m.group(5)),
            second=int(m.group(6)),
            tzinfo=local_timezone,
        )

        _LOGGER.debug(f"LastUpdated {last_updated} - Timezone {local_timezone}")

        self.last_updated = last_updated

    def set_engine_type(self):
        if "evStatus" in self.vehicle_data[
            "vehicleStatus"
        ] and self.token.vehicle_model.endswith(" EV"):
            self.engine_type = VEHICLE_ENGINE_TYPE.EV
        else:
            if (
                "evStatus" in self.vehicle_data["vehicleStatus"]
                and "lowFuelLight" in self.vehicle_data["vehicleStatus"]
            ):
                self.engine_type = VEHICLE_ENGINE_TYPE.PHEV
            else:
                if "evStatus" in self.vehicle_data["vehicleStatus"]:
                    self.engine_type = VEHICLE_ENGINE_TYPE.EV
                else:
                    self.engine_type = VEHICLE_ENGINE_TYPE.IC
        _LOGGER.debug(f"Engine type set {self.engine_type}")

    def get_child_value(self, key):
        value = self.vehicle_data
        for x in key.split("."):
            try:
                value = value[x]
            except:
                try:
                    value = value[int(x)]
                except:
                    value = None
        return value
